import osmnx
import pytest

from maskmypy import tools


def test_atlas_mask_snap(points_small):
    unsnapped = points_small
    snapped = tools.snap_to_streets(points_small)

    with pytest.raises(osmnx._errors.InsufficientResponseError):
        unsnapped_point = unsnapped.to_crs(epsg=4326).iloc[0].geometry
        osmnx.features.features_from_point(
            (unsnapped_point.y, unsnapped_point.x), tags={"highway": True}, dist=2
        )

    snapped_point = snapped.to_crs(epsg=4326).iloc[0].geometry
    osmnx.features.features_from_point(
        (snapped_point.y, snapped_point.x), tags={"highway": True}, dist=2
    )
    assert snapped.crs == unsnapped.crs
    assert snapped.iloc[0].geometry != unsnapped.iloc[0].geometry


def test_unmasked_points(points):
    masked = points.copy()
    i = 5

    masked.loc[i:, "geometry"] = masked.loc[i:, masked.geometry.name].translate(1, 0, 0)
    with pytest.warns(UserWarning):
        masked = tools._mark_unmasked_points(points, masked)

    assert all(masked.loc[i:, "UNMASKED"] == 0)
    assert all(masked.loc[: i - 1, "UNMASKED"] == 1)
    assert masked["UNMASKED"].sum() == i



from copy import deepcopy
from dataclasses import dataclass

from geopandas import GeoDataFrame
from networkx import single_source_dijkstra_path_length
from numpy import random
from osmnx import graph_to_gdfs
from osmnx.distance import add_edge_lengths, nearest_nodes
from osmnx.graph import graph_from_bbox
from osmnx.utils_graph import remove_isolated_nodes
from shapely import Point

from .. import tools


def street(
    gdf: GeoDataFrame,
    low: int,
    high: int,
    max_length: float = 1000,
    seed: int = None,
    padding: float = 0.2,
) -> GeoDataFrame:
    """
    Apply street masking to a GeoDataFrame, displacing points along the OpenStreetMap street
    network. This helps account for variations in population density, and reduces the likelihood
    of false attribution as points are always displaced to the street network. Each point is
    snapped to the nearest node on the network, then displaced along the surround network between
    `low` and `high` nodes away.

    Example
    -------
    ```python
    from maskmypy import street

    masked = street(
        gdf=sensitive_points,
        min=20,
        max=30
    )
    ```

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame containing sensitive points.
    low : int
        Minimum number of nodes along the OSM street network to traverse.
    high : int
        Maximum number of nodes along the OSM street network to traverse.
    max_length : float
        When locating the closest node to each point on the street network, MaskMyPy verifies
        that its immediate neighbours are no more than `max_length` away, in meters. This prevents
        extremely large masking distances, such as those caused by long highways.
    seed : int
        Used to seed the random number generator so that masked datasets are reproducible.
        Randomly generated if left undefined.
    padding : float
        OSM network data is retrieved based on the bounding box of the sensitive GeoDataFrame.
        Padding is used to expand this bounding box slightly to reduce unwanted edge-effects.
        A value of `0.2` would add 20% of the x and y extent to *each side* of the bounding box.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing masked points.
    """
    _gdf = gdf.copy()
    _validate_street(_gdf, low, high)

    seed = tools.gen_seed() if not seed else seed

    args = locals()
    del args["gdf"]

    masked_gdf = _Street(**args).run()

    return masked_gdf


def _validate_street(gdf, low, high):
    tools._validate_geom_type(gdf, "Point")

    if low >= high:
        raise ValueError("Low value must be less than high value.")


@dataclass
class _Street:
    _gdf: GeoDataFrame
    low: int
    high: int
    max_length: float
    seed: int
    padding: float

    def __post_init__(self) -> None:
        self._rng = random.default_rng(seed=self.seed)
        self.crs = self._gdf.crs
        self._gdf = self._gdf.to_crs(epsg=4326)
        self._get_osm()

    def _get_osm(self) -> None:
        bbox = tools._pad(self._gdf.total_bounds, self.padding)
        self.graph = add_edge_lengths(
            remove_isolated_nodes(  # This will be deprecated in OSMNX 2.0
                graph_from_bbox(
                    bbox=(bbox[3], bbox[1], bbox[2], bbox[0]),
                    network_type="drive",
                    truncate_by_edge=True,
                ),
                warn=False,
            )
        )
        self.graph_gdf = graph_to_gdfs(self.graph)
        self.graph_tmp = deepcopy(self.graph)

    def _mask_point(self, point: Point) -> Point:
        nearest_node_id = self._nearest_node_with_neighbors(point)
        return self._mask_point_from_node_id(nearest_node_id)

    def _nearest_node_with_neighbors(self, point: Point) -> int:
        neighbor_count = 0
        while neighbor_count < 1:
            node_id = nearest_nodes(self.graph_tmp, point.x, point.y)
            neighbors = list(self.graph_tmp.neighbors(node_id))
            for neighbor in neighbors:
                length = self.graph_tmp.get_edge_data(node_id, neighbor)[0]["length"]
                if length > self.max_length:
                    neighbors = [n for n in neighbors if n != neighbor]
            neighbor_count = len(neighbors)

            if neighbor_count == 0:
                self.graph_tmp.remove_node(node_id)
        return node_id

    def _mask_point_from_node_id(self, node_id: int) -> Point:
        node_count = 0
        target_node_count = self._rng.integers(self.low, self.high, endpoint=False)
        _max_length = self.max_length

        # Search for surrounding nodes until number of nodes found equals or exceeds target_node_count
        while node_count < target_node_count:
            paths = single_source_dijkstra_path_length(
                G=self.graph, source=node_id, cutoff=_max_length, weight="length"
            )
            node_count = len(paths)
            _max_length = _max_length * 2

        # Calculate the total distance to the closest nodes within the target_node_count
        i = 0
        distances = []
        nodes = []
        total_distance = 0
        for key, value in paths.items():
            nodes.append(key)
            distances.append(value)
            total_distance += value
            i += 1
            if i == target_node_count:
                break

        # Target distance is the average distance to nodes within the target_node_count
        target_distance = total_distance / target_node_count

        # Find the node closest to the candidate distance
        target_node = nodes[
            distances.index(min(distances, key=lambda x: abs(x - target_distance)))
        ]

        # Return the geometry of the target node
        return self.graph_gdf[0].at[target_node, self.graph_gdf[0].geometry.name]

    def run(self) -> GeoDataFrame:
        self._gdf[self._gdf.geometry.name] = self._gdf[self._gdf.geometry.name].apply(
            self._mask_point
        )
        self._gdf = self._gdf.to_crs(self.crs)
        return self._gdf



import warnings
from hashlib import sha256
from random import SystemRandom

from geopandas import GeoDataFrame
from numpy import random
from osmnx import graph_to_gdfs
from osmnx.distance import nearest_nodes
from osmnx.graph import graph_from_bbox
from osmnx.projection import project_graph
from osmnx.utils_graph import remove_isolated_nodes
from pandas.util import hash_pandas_object
from pyproj.crs.crs import CRS


def checksum(gdf: GeoDataFrame) -> str:
    """
    Calculate SHA256 checksum of a GeoDataFrame and return the first 8 characters.
    Two completely identical GeoDataFrames will always return the exact same value,
    whereas two similar, but not completely identical GeoDataFrames will return
    entirely different values.

    Parameters
    ----------
    gdf : GeoDataFrame
        Any valid GeoDataFrame.

    Returns
    -------
    str
        The first 8 characters of the SHA256 checksum of the input GeoDataFrame.
    """
    return sha256(bytearray(hash_pandas_object(gdf).values)).hexdigest()[0:8]


def gen_rng(seed: int = None) -> object:
    """
    Create a seeded numpy default_rng() object.

    Parameters
    ----------
    seed : int
        An integer used to seed the random number generator. A seed is randomly
        generated using gen_seed() if one is not provided.
    Returns
    -------
    object
        numpy.default_rng()
    """
    if not seed:
        seed = gen_seed()
    return random.default_rng(seed=seed)


def gen_seed() -> int:
    """
    Generate a 16-digit random integer to seed random number generators.

    Returns
    -------
    int
        A 16 digit random integer.
    """

    return int(SystemRandom().random() * (10**16))


def snap_to_streets(gdf: GeoDataFrame) -> GeoDataFrame:
    """
    Relocates each point of a GeoDataFrame to the nearest node on the OpenStreetMap driving
    network. Performing this on masked datasets may reduce the chances of false attribution,
    and may provide an additional layer of obfuscation.

    This is *not* an alternative to masking.

    Parameters
    ----------
    gdf : GeoDataFrame
        A GeoDataFrame containing point data.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing points that have been snapped to street nodes.
    """
    snapped_gdf = gdf.copy()
    bbox = gdf.to_crs(epsg=4326).total_bounds
    graph = remove_isolated_nodes(
        graph_from_bbox(
            bbox=(bbox[3], bbox[1], bbox[2], bbox[0]),
            network_type="drive",
            truncate_by_edge=True,
        ),
        warn=False,
    )
    graph = project_graph(graph, to_crs=gdf.crs)
    node_gdf = graph_to_gdfs(graph)[0]

    snapped_gdf[snapped_gdf.geometry.name] = snapped_gdf[snapped_gdf.geometry.name].apply(
        lambda geom: node_gdf.at[nearest_nodes(graph, geom.x, geom.y), node_gdf.geometry.name]
    )

    return snapped_gdf


def _mark_unmasked_points(sensitive: GeoDataFrame, masked: GeoDataFrame):
    geom_col_idx = sensitive.columns.get_loc(sensitive.geometry.name)
    masked["UNMASKED"] = masked.apply(
        lambda x: (1 if x[masked.geometry.name] == sensitive.iat[x.name, geom_col_idx] else 0),
        axis=1,
    )
    unmasked_count = masked["UNMASKED"].sum()
    if unmasked_count > 0:
        warnings.warn(
            f"{unmasked_count} points could not be masked. Adding 'UNMASKED' column to mark unmasked points."
        )
    return masked


def _crop(gdf: GeoDataFrame, bbox: list[float], padding: float) -> GeoDataFrame:
    bbox = _pad(bbox, padding)
    return gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]]


def _pad(bbox: list[float], padding: float) -> list:
    pad_x = (bbox[2] - bbox[0]) * padding
    pad_y = (bbox[3] - bbox[1]) * padding
    bbox[0] = bbox[0] - pad_x
    bbox[1] = bbox[1] - pad_y
    bbox[2] = bbox[2] + pad_x
    bbox[3] = bbox[3] + pad_y
    return bbox


def _validate_geom_type(gdf: GeoDataFrame, *type_as_string: str) -> bool:
    geom_types = {True if geom_type in type_as_string else False for geom_type in gdf.geom_type}
    if False in geom_types:
        raise ValueError(f"GeoDataFrame contains geometry types other than {type_as_string}.")
    return True


def _validate_crs(*crs: CRS, custom_message: str = None) -> bool:
    default_message = (
        "CRS mismatch. Ensure the coordinate reference systems of all input layers match."
    )
    message = default_message if not custom_message else custom_message
    if len(set(crs)) != 1:
        raise ValueError(message)
    else:
        return True



import os
import shutil

import geopandas as gpd
import pytest


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture()
def tmpdir():
    os.makedirs("./tmp/", exist_ok=True)
    os.chdir("./tmp")
    yield
    os.chdir("../")
    shutil.rmtree("./tmp")


POINTS = gpd.read_file("tests/data/points.geojson").to_crs(epsg=26910)
ADDRESS = gpd.read_file("tests/data/addresses.geojson").to_crs(epsg=26910)
CONTAINER = gpd.read_file("tests/data/boundary.geojson").to_crs(epsg=26910)


@pytest.fixture
def points():
    return POINTS.copy()


@pytest.fixture
def points_small():
    points = POINTS.copy()
    return points.clip(points.iloc[0].geometry.buffer(1500))


@pytest.fixture
def address():
    return ADDRESS.copy()


@pytest.fixture
def container():
    return CONTAINER.copy()


# @pytest.fixture
# def atlas(points, addresses, tmpdir):
#     return Atlas(name="test_atlas", directory="./tmp/", input=points, population=addresses)


# @pytest.fixture
# def atlas_contained(points, container, tmpdir):
#     return Atlas(name="test_atlas", container=container, directory="./tmp/", input=points)



import inspect
import json
import pprint
import tracemalloc
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from timeit import default_timer
from typing import Callable

import matplotlib.pyplot as plt
from geopandas import GeoDataFrame
from pandas import DataFrame, Series, concat

from . import analysis, masks, tools


@dataclass
class Atlas:
    """
    A class for quickly performing and evaluating geographic masks.

    Example
    -------
    ```python
    from maskmypy import Atlas, donut, locationswap

    atlas = Atlas(sensitive=some_points, population=some_addresses)
    atlas.mask(donut, low=50, high=500)
    atlas.mask(locationswap, low=50, high=500, address=some_addresses)
    atlas.as_df()
    ```

    Attributes
    ----------
    sensitive : GeoDataFrame
        A GeoDataFrame containing sensitive points.
    population : GeoDataFrame
        A GeoDataFrame containing population information, such as address points or polygon
        with population counts.
    population_column : str
        If the population layer is based on polygons, the name of the column containing population
        counts.
    candidates : list[]
        A list of existing masked candidates, if any.
    """

    sensitive: GeoDataFrame
    population: GeoDataFrame = None
    population_column: str = "pop"
    candidates: list = field(default_factory=list)

    def __post_init__(self):
        self.layers = {}
        if isinstance(self.population, GeoDataFrame):
            tools._validate_crs(self.sensitive.crs, self.population.crs)

    def __getitem__(self, idx):
        return self.candidates[idx]

    def __setitem__(self, idx, val):
        self.candidates[idx] = val

    def __len__(self):
        return len(self.candidates)

    def add_layers(self, *gdf: GeoDataFrame):
        """
        Add GeoDataFrames to the layer store (`Atlas.layers`).

        When regenerating masked GeoDataFrames using `Atlas.gen_gdf()`, any context layers
        that were used in creating the associated candidate must be present in the layer store.
        If they are, they will be automatically found and used as needed.

        Note that layers are stored according to their checksum value (see
        `maskmypy.tools.checksum()`) to provide both deduplication and integrity
        checking.

        Parameters
        ----------
        gdf : GeoDataFrame
            GeoDataFrames to be added to the layer store.
        """
        for x in gdf:
            tools._validate_crs(self.sensitive.crs, x.crs)
            self.layers[tools.checksum(x)] = x

    def mask(
        self,
        mask_func: Callable,
        keep_gdf: bool = False,
        keep_candidate: bool = True,
        skip_slow_evaluators: bool = True,
        measure_execution_time: bool = True,
        measure_peak_memory: bool = False,
        **kwargs,
    ):
        """
        Execute a given mask, analyze the result, and add it to the Atlas.

        Parameters
        ----------
        mask_func : GeoDataFrame
            A masking function to apply to the sensitive point dataset. If using a custom mask,
            it must take the sensitive GeoDataFrame as its first argument, all other arguments as
            keyword arguments, and must return a GeoDataFrame containing the results.
        keep_gdf : bool
            If `False`, the resulting GeoDataFrame will be analyzed and then dropped to save memory.
            Use `gen_gdf` to regenerate the GeoDataFrame.
        keep_candidate : bool
            If `True`, a dictionary containing mask parameters and analysis results are added to
            the candidate list (`Atlas.candidates`, or `Atlas[index]`).
        skip_slow_evaluators : bool
            If `True`, skips any analyses that are known to be slow during mask result
            evaluation. See maskmypy.analysis.evaluate() for more information.
        measure_execution_time : bool
            If `True`, measures the execution time of the mask function and adds it to the
            candidate statistics. Mutually exclusive with `measure_peak_memory`
        measure_peak_memory : bool
            If `True`, will profile memory usage while the mask function is being applied,
            and will add the value in MB to the candidate statistics. Note that the reported
            value represents *additional* memory used by the mask, and does not include existing
            allocations. Mutually exclusive with `measure_peak_memory`.

            Warning: this can significantly slow down execution time.

        """
        if measure_execution_time and measure_peak_memory:
            raise ValueError(
                "`measure_execution_time` and `measure_peak_memory` cannot both be true."
            )

        candidate = {
            "mask": mask_func.__name__,
            "kwargs": self._hydrate_mask_kwargs(**kwargs),
        }

        if "seed" in inspect.getfullargspec(mask_func).args and "seed" not in candidate["kwargs"]:
            candidate["kwargs"]["seed"] = tools.gen_seed()

        if measure_execution_time:
            time_start = default_timer()
        elif measure_peak_memory:
            tracemalloc.start()

        gdf = mask_func(self.sensitive, **candidate["kwargs"])

        if measure_execution_time:
            execution_time = default_timer() - time_start
        elif measure_peak_memory:
            _, mem_peak = tracemalloc.get_traced_memory()
            tracemalloc.stop()
            mem_peak_mb = mem_peak / 1024 / 1024

        candidate["checksum"] = tools.checksum(gdf)
        candidate["kwargs"] = self._dehydrate_mask_kwargs(**candidate["kwargs"])
        candidate["stats"] = analysis.evaluate(
            sensitive_gdf=self.sensitive,
            candidate_gdf=gdf,
            population_gdf=self.population,
            population_column=self.population_column,
            skip_slow=skip_slow_evaluators,
        )

        if "UNMASKED" in gdf.columns:
            candidate["stats"]["UNMASKED_POINTS"] = gdf["UNMASKED"].sum()

        if measure_execution_time:
            candidate["stats"]["execution_time"] = round(execution_time, 3)
        elif measure_peak_memory:
            candidate["stats"]["memory_peak_mb"] = round(mem_peak_mb, 3)

        if keep_gdf:
            self.layers[candidate["checksum"]] = gdf
        else:
            del gdf

        if keep_candidate:
            self.candidates.append(candidate)

        return candidate

    def gen_gdf(
        self,
        idx: int = None,
        checksum: str = None,
        keep: bool = False,
        custom_mask: Callable = None,
    ):
        """
        Regenerates the GeoDataFrame for a given candidate based on either its position in the
        `Atlas.candidates` list or its checksum.

        Parameters
        ----------
        idx : int
            Index of the candidate in `Atlas.candidates` to regenerate a GeoDataFrame for.
        checksum : str
            Checksum of the candidate in `Atlas.candidates` to regenerate a GeoDataFrame for.
        keep : bool
            If `True`, return the masked GeoDataFrame and store it in `Atlas.layers` for future
            use so it does not need to be regenerated.
        custom_mask : Callable
            If the candidate was generated using a custom masking function from outside MaskMyPy,
            provide the function here.

        """
        if (idx is None and checksum is None) or (idx is not None and checksum is not None):
            raise ValueError(f"Must specify either idx or checksum.")

        checksum_before = checksum if checksum else self.candidates[idx]["checksum"]

        # Check if layer is already in the store.
        if isinstance(self.layers.get(checksum_before, None), GeoDataFrame):
            return self.layers[checksum_before]

        try:
            candidate = next(
                cand for cand in self.candidates if cand["checksum"] == checksum_before
            )
        except:
            raise ValueError(f"Could not locate candidate with checksum '{checksum_before}'")

        mask_func = custom_mask or getattr(masks, candidate["mask"])

        candidate_after = self.mask(
            mask_func, keep_candidate=False, keep_gdf=True, **candidate["kwargs"]
        )

        checksum_after = candidate_after.get("checksum")
        if checksum_before != checksum_after:
            raise ValueError(
                f"Checksum of masked GeoDataFrame ({checksum_after}) does not match that which is on record for this candidate ({checksum_before}). Did any input layers get modified?"
            )

        gdf = self.layers[checksum_after]

        if not keep:
            del self.layers[checksum_after]

        return gdf

    def sort(self, by: str, desc: bool = False):
        """
        Sorts the list of candidates (`Atlas.candidates`) based on a given statistic.

        Example:
        ```
        # Sort candidate list in ascending order based on maximum displacement distance.
        atlas.sort(by="displacement_max")

        # Sort candidate list in descending order based on minimum k-anonymity.
        atlas.sort(by="k_min", desc=True)
        ```

        Parameters
        ----------
        by : str
            Name of the statistic to sort by.
        desc : bool
            If `True`, sort in descending order.

        """
        if by in self.candidates[0]["stats"].keys():
            self.candidates.sort(key=lambda x: x["stats"][by], reverse=desc)
        else:
            raise ValueError(f"Could not find '{by}' in candidate statistics.")

    def prune(self, by: str, min: float, max: float):
        """
        Prune candidates based on a given statistic. If the value for that attribute is less than
        `min` or greater than `max` (both inclusive), drop the candidate.

        Example:
        ```
        # Prune any candidates with a minimum displacement distance below 50 and above 500.
        atlas.prune(by="displacement_min", min=50, max=500)

        # Prune any candidates with minimum k-anonymity values below 10 and above 50.
        atlas.prune(by="k_min", min=10, max=50)
        ```

        Parameters
        ----------
        by : str
            Name of the candidate statistic to prune by.
        min : float
            Minimum value of the statistic. If below `min`, the candidate is pruned from the
            candidates list. If the statistic is equal to or greater than `min` but not
            greater than `max` it is kept in the list.
        max : float
            Maximum value of the statistic. If above `max`, the candidate is pruned from the
            candidates list. If the statistic is equal to or less than `max` but not less
            than `min` it is kept in the list.
        """
        if by in self.candidates[0]["stats"].keys():
            self.candidates = [
                c for c in self.candidates if c["stats"][by] >= min and c["stats"][by] <= max
            ]
        else:
            raise ValueError(f"Could not find '{by}' in candidate statistics.")

    def to_json(self, file: Path):
        """
        Saves candidates to a JSON file. As long as the input GeoDataFrames are
        also preserved by the user*, this JSON file can be used to later reconstruct
        the atlas using `Atlas.from_json()`, including all resulting candidate GeoDataFrames.

        * Warning: if Street masking is used, there is a chance that a candidate will not be able
        to be regenerated if OpenStreetMap data changes. This will be addressed in a future version
        of MaskMyPy.

        Parameters
        ----------
        file : Path
            File path indicating where the JSON file should be saved.
        """
        with open(file, "w") as f:
            json.dump(self.candidates, f)

    @classmethod
    def from_json(
        cls,
        sensitive: GeoDataFrame,
        candidate_json: Path,
        population: GeoDataFrame = None,
        population_column: str = "pop",
        layers: list = None,
    ):
        """
        Recreate an Atlas from a candidate JSON file previously generated using `Atlas.to_json()`
        as well as the original GeoDataFrames. Masked GeoDataFrames can then be regenerated using
        `Atlas.gen_gdf()`.

        * Warning: if Street masking is used, there is a chance that a candidate will not be able
        to be regenerated if OpenStreetMap data changes. This will be addressed in a future version
        of MaskMyPy.

        Parameters
        ----------
        sensitive : GeoDataFrame
            The original sensitive point layer.
        candidate_json : Path
            Path to a candidate JSON file previously generated using `Atlas.to_json()`.
        population : GeoDataFrame
            The original population layer, if one was specified.
        population_column : str
            If a polygon-based population layer was used, the name of the population column.
        layers : List[GeoDataFrame]
            A list of additional GeoDataFrames used in the original Atlas. For instance,
            any containers used during donut masking.
        """
        with open(candidate_json) as f:
            candidates = json.load(f)

        atlas = cls(
            sensitive=sensitive,
            candidates=candidates,
            population=population,
            population_column=population_column,
        )
        if layers:
            atlas.add_layers(*layers)
        return atlas

    def as_df(self):
        """
        Return a pandas DataFrame describing each candidate.
        """
        df = DataFrame(data=self.candidates)
        df = concat([df.drop(["kwargs"], axis=1), df["kwargs"].apply(Series)], axis=1)
        df = concat([df.drop(["stats"], axis=1), df["stats"].apply(Series)], axis=1)
        return df

    def scatter(self, a: str, b: str):
        """
        Return a scatter plot of candidates across two given statistics.

        Parameters
        ----------
        a : string
            Name of the candidate statistic to plot.
        b : string
            Name of the candidate statistic to plot.
        """
        df = self.as_df()
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.scatter(df[a], df[b], c="#1f77b4")
        ax.set_xlabel(a)
        ax.set_ylabel(b)
        for i, label in enumerate(df["checksum"]):
            ax.annotate(label, (df.loc[i, a], df.loc[i, b]))
        return fig

    def _hydrate_mask_kwargs(self, **mask_kwargs: dict) -> dict:
        """
        Find any keyword arguments that contain context layer checksums and
        attempt to restore the layer from `Atlas.layers`.
        """
        for key, value in mask_kwargs.items():
            if isinstance(value, str) and value.startswith("context_"):
                checksum = value.split("_")[1]
                try:
                    mask_kwargs[key] = self.layers[checksum]
                except KeyError as e:
                    raise KeyError(
                        f"Error: cannot find context layer for '{key}, {checksum}', \
                        try loading it first using Atlas.add_layers(). {e}"
                    )
        return mask_kwargs

    def _dehydrate_mask_kwargs(self, **mask_kwargs: dict) -> dict:
        """
        Search mask kwargs for any GeoDataFrames and replace them with their checksums.
        """
        for key, value in mask_kwargs.items():
            if isinstance(value, GeoDataFrame):
                self.add_layers(value)
                mask_kwargs[key] = "_".join(["context", tools.checksum(value)])
        return mask_kwargs



from dataclasses import dataclass

from geopandas import GeoDataFrame, GeoSeries
from shapely import Point, voronoi_polygons
from shapely.ops import nearest_points

from .. import tools


def voronoi(gdf: GeoDataFrame, snap_to_streets: bool = False) -> GeoDataFrame:
    """
    Apply voronoi masking to a GeoDataFrame, displacing points to the nearest edges of a vornoi
    diagram. Note: because voronoi masking lacks any level of randomization, snapping to streets
    is recommended for this mask to provide another level of obfuscation.

    Example
    -------
    ```python
    from maskmypy import voronoi

    masked = voronoi(
        gdf=sensitive_points,
        snap_to_streets=True
    )
    ```

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame containing sensitive points.
    snap_to_streets : bool
        If True, points are snapped to the nearest node on the OSM street network after masking.
        This can reduce the chance of false-attribution.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing masked points.
    """
    _gdf = gdf.copy()
    _validate_voronoi(gdf)

    args = locals()
    del args["snap_to_streets"]
    del args["gdf"]

    masked_gdf = _Voronoi(**args).run()

    if snap_to_streets:
        masked_gdf = tools.snap_to_streets(masked_gdf)

    return masked_gdf


def _validate_voronoi(gdf):
    tools._validate_geom_type(gdf, "Point")


@dataclass
class _Voronoi:
    _gdf: GeoDataFrame

    def _generate_voronoi_polygons(self) -> GeoSeries:
        points = self._gdf.dissolve()
        return voronoi_polygons(points.geometry, only_edges=True)

    def _mask_point(self, point: Point) -> Point:
        return nearest_points(point, self.voronoi)[1][0]

    def run(self) -> GeoDataFrame:
        self.voronoi = self._generate_voronoi_polygons()
        self._gdf[self._gdf.geometry.name] = self._gdf[self._gdf.geometry.name].apply(
            self._mask_point
        )

        return self._gdf



from dataclasses import dataclass

from geopandas import GeoDataFrame
from numpy import random
from shapely import Point

from .. import tools


def locationswap(
    gdf: GeoDataFrame,
    low: float,
    high: float,
    address: GeoDataFrame,
    seed: int = None,
    snap_to_streets: bool = False,
):
    """
    Applies location swapping to a GeoDataFrame, displacing points to a randomly selected address
    that is between a minimum and maximum distance away from the original point. While address
    data is the most common data type used to provide eligible swap locations, other point-based
    datasets may be used.

    Note: If a sensitive point has no address points within range, the point is displaced to (0,0).

    Example
    -------
    ```python
    from maskmypy import locationswap

    masked = locationswap(
        gdf=sensitive_points,
        low=50,
        high=500,
        address=address_points
    )
    ```

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame containing sensitive points.
    low : float
        Minimum distance to displace points. Unit must match that of the `gdf` CRS.
    high : float
        Maximum displacement to displace points. Unit must match that of the `gdf` CRS.
    address : GeoDataFrame
        GeoDataFrame containing points that sensitive locations may be swapped to.
        While addresses are most common, other point-based data may be used as well.
    seed : int
        Used to seed the random number generator so that masked datasets are reproducible.
        Randomly generated if left undefined.
    snap_to_streets : bool
        If True, points are snapped to the nearest node on the OSM street network after masking.
        This can reduce the chance of false-attribution.
    """

    _gdf = gdf.copy()
    _validate_locationswap(_gdf, low, high, address)

    seed = tools.gen_seed() if not seed else seed

    args = locals()
    del args["snap_to_streets"]
    del args["gdf"]

    mask = _LocationSwap(**args)
    masked_gdf = mask.run()

    if mask._unmasked_points:
        masked_gdf = tools._mark_unmasked_points(gdf, masked_gdf)

    if snap_to_streets:
        masked_gdf = tools.snap_to_streets(masked_gdf)

    return masked_gdf


def _validate_locationswap(gdf, low, high, address):
    tools._validate_geom_type(gdf, "Point")
    tools._validate_geom_type(address, "Point")
    tools._validate_crs(gdf.crs, address.crs)

    if low >= high:
        raise ValueError("Minimum displacement distance is larger than or equal to maximum.")


@dataclass
class _LocationSwap:
    _gdf: GeoDataFrame
    low: float
    high: float
    address: GeoDataFrame
    seed: int = None

    def __post_init__(self) -> None:
        # Initialize random number generator
        self._rng = random.default_rng(seed=self.seed)
        self._unmasked_points = False

    def _mask_point(self, point: Point) -> Point:
        min_buffer = point.buffer(self.low)
        max_buffer = point.buffer(self.high)

        excluded_locations = self.address.intersects(min_buffer)
        excluded_locations = set(excluded_locations[excluded_locations].index)

        included_locations = self.address.intersects(max_buffer)
        included_locations = set(included_locations[included_locations].index)

        included_locations = list(included_locations.difference(excluded_locations))
        if len(included_locations) > 0:
            return self.address.iloc[self._rng.choice(included_locations)].geometry
        else:
            self._unmasked_points = True
            return point

    def run(self) -> GeoDataFrame:
        self._gdf[self._gdf.geometry.name] = self._gdf[self._gdf.geometry.name].apply(
            self._mask_point
        )
        return self._gdf



import statistics
import time

import pytest

from maskmypy import Atlas, analysis, donut, tools, voronoi


def test_atlas_mask(points):
    atlas = Atlas(points)
    atlas.mask(donut, low=50, high=500)
    assert len(atlas[:]) == 1
    assert atlas[0]["checksum"] != tools.checksum(points)


def test_atlas_as_df(points):
    atlas = Atlas(points)
    atlas.mask(donut, low=50, high=500)
    df = atlas.as_df()
    assert df.iloc[0]["high"] == 500
    assert df.iloc[0]["mask"] == "donut"


def test_atlas_restore_from_json(points_small):
    points = points_small
    atlas = Atlas(points)

    atlas.mask(donut, low=10, high=100)
    atlas.mask(donut, low=50, high=500, snap_to_streets=True)

    check_1a = atlas[0]["checksum"]
    check_2a = atlas[1]["checksum"]

    atlas.to_json("/tmp/tmp_test.json")
    del atlas

    # Test by index value
    atlas2 = Atlas.from_json(points, "/tmp/tmp_test.json")

    gdf_0 = atlas2.gen_gdf(0)
    check_1b = tools.checksum(gdf_0)
    assert check_1a == check_1b

    gdf_1 = atlas2.gen_gdf(1)
    check_2b = tools.checksum(gdf_1)
    assert check_2a == check_2b

    # Test by checksum value
    atlas3 = Atlas.from_json(points, "/tmp/tmp_test.json")

    gdf_0 = atlas3.gen_gdf(checksum=check_1a)
    check_1c = tools.checksum(gdf_0)
    assert check_1a == check_1c

    gdf_1 = atlas3.gen_gdf(checksum=check_2a)
    check_2c = tools.checksum(gdf_1)
    assert check_2a == check_2c

    with pytest.raises(ValueError):
        atlas3.gen_gdf(checksum="aaaaaa")


def test_atlas_context_hydration(points, container):
    atlas = Atlas(points)
    atlas.mask(donut, container=container, low=50, high=500)
    atlas.to_json("/tmp/tmp_test.json")
    del atlas

    atlas2 = Atlas.from_json(points, "/tmp/tmp_test.json")
    with pytest.raises(KeyError):
        atlas2.gen_gdf(0)

    atlas2.add_layers(container)
    atlas2.gen_gdf(0)
    del atlas2

    atlas3 = Atlas.from_json(points, "/tmp/tmp_test.json", layers=[container])
    atlas3.gen_gdf(0)


def test_atlas_sort(points):
    atlas = Atlas(points)
    atlas.mask(donut, low=300, high=399)
    atlas.mask(donut, low=200, high=299)
    atlas.mask(donut, low=100, high=199)

    assert (
        atlas[0]["stats"]["displacement_mean"]
        > atlas[1]["stats"]["displacement_mean"]
        > atlas[2]["stats"]["displacement_mean"]
    )

    atlas.sort(by="displacement_mean")
    assert (
        atlas[0]["stats"]["displacement_mean"]
        < atlas[1]["stats"]["displacement_mean"]
        < atlas[2]["stats"]["displacement_mean"]
    )

    atlas.sort(by="displacement_mean", desc=True)
    assert (
        atlas[0]["stats"]["displacement_mean"]
        > atlas[1]["stats"]["displacement_mean"]
        > atlas[2]["stats"]["displacement_mean"]
    )


def test_displacement(points):
    masked_gdf = donut(points, 50, 500)
    displacement_gdf = analysis.displacement(points, masked_gdf)
    assert "_distance" not in points.columns
    assert "_distance" not in masked_gdf.columns
    assert "_distance" in displacement_gdf.columns


def test_evaluate(points, address):
    atlas = Atlas(points, population=address)
    atlas.mask(donut, low=100, high=199, skip_slow_evaluators=False)
    atlas.mask(donut, low=300, high=399, skip_slow_evaluators=False)
    assert atlas[0]["stats"]["displacement_min"] < atlas[1]["stats"]["displacement_max"]
    assert atlas[0]["stats"]["k_satisfaction_50"] < atlas[0]["stats"]["k_satisfaction_5"]


def test_ripley(points):
    atlas = Atlas(points)
    lows = []
    for i in range(0, 4):
        atlas.mask(donut, low=1, high=100, skip_slow_evaluators=False)
        lows.append(atlas[i]["stats"]["ripley_rmse"])

    highs = []
    for i in range(4, 7):
        atlas.mask(donut, low=100, high=200, skip_slow_evaluators=False)
        highs.append(atlas[i]["stats"]["ripley_rmse"])

    assert (statistics.mean(lows)) < (statistics.mean(highs))


def test_atlas_prune(points, address):
    atlas = Atlas(points, population=address)
    atlas.mask(donut, low=300, high=399)
    atlas.mask(donut, low=200, high=299)
    atlas.mask(donut, low=100, high=199)

    atlas.prune(by="displacement_min", min=200, max=9999)
    assert len(atlas.candidates) == 2

    atlas.prune(by="displacement_min", min=0, max=299)
    assert len(atlas.candidates) == 1


def test_atlas_crs_mismatch(points, address):
    address = address.to_crs(epsg=4326)
    with pytest.raises(ValueError):
        atlas = Atlas(points, population=address)


def test_execution_time(points):
    atlas = Atlas(points)

    def mask_mock(sensitive, seed):
        time.sleep(0.1)
        return sensitive

    atlas.mask(mask_mock, measure_execution_time=True)
    assert round(atlas[0]["stats"]["execution_time"], 1) == 0.1


def test_peak_memory(points):
    points = points[0:2]
    atlas = Atlas(points)

    def mask_mock(sensitive, memory_mb, seed):
        peakmemory_size = memory_mb * 1024 * 1024  # 500MB
        byte_array = bytearray(peakmemory_size)
        return sensitive

    atlas.mask(mask_mock, memory_mb=1, measure_peak_memory=True, measure_execution_time=False)
    assert round(atlas[0]["stats"]["memory_peak_mb"]) == 1

    atlas.mask(mask_mock, memory_mb=100, measure_peak_memory=True, measure_execution_time=False)
    assert round(atlas[1]["stats"]["memory_peak_mb"]) == 100

    atlas.mask(mask_mock, memory_mb=10, measure_peak_memory=True, measure_execution_time=False)
    assert round(atlas[2]["stats"]["memory_peak_mb"]) == 10


def test_seed_not_in_voronoi_candidate(points):
    atlas = Atlas(points)
    atlas.mask(voronoi, snap_to_streets=False)
    assert "seed" not in atlas[0]["kwargs"]
    atlas.gen_gdf(idx=0)


def test_memory_and_speed_are_exclusive(points):
    atlas = Atlas(points)
    with pytest.raises(ValueError):
        atlas.mask(donut, low=1, high=2, measure_peak_memory=True, measure_execution_time=True)



import os

import geopandas as gpd
from numpy import floor
from shapely.geometry import Point, Polygon

from maskmypy import analysis, donut


def test_k_satisfaction(points, address):
    masked = donut(points, 100, 500)
    masked_k = analysis.k_anonymity(points, masked, address)
    k_sat_1 = analysis.k_satisfaction(masked_k, 1)
    k_sat_50 = analysis.k_satisfaction(masked_k, 50)
    k_sat_999 = analysis.k_satisfaction(masked_k, 999)
    assert k_sat_1 > 0.9
    assert 0.1 < k_sat_50 < 0.9
    assert k_sat_999 < 0.1


def test_k_summary(points, address):
    masked = donut(points, 100, 500)
    masked_k = analysis.k_anonymity(points, masked, address)
    k_sum = analysis.summarize_k(masked_k)
    assert k_sum["k_min"] < k_sum["k_mean"] < k_sum["k_max"]


def test_displacement(points):
    masked_points = points.copy()
    masked_points["geometry"] = masked_points.geometry.translate(50, 0, 0)
    displacement = analysis.summarize_displacement(analysis.displacement(points, masked_points))
    assert displacement["displacement_min"] == 50
    assert displacement["displacement_max"] == 50
    assert displacement["displacement_med"] == 50
    assert displacement["displacement_mean"] == 50


def test_estimate_k_address():
    addr_points = {
        "geometry": [
            Point(0, 0),
            Point(1, 0),
            Point(2, 0),
            Point(3, 0),
            Point(4, 0),
            Point(5, 0),
            Point(7, 0),
        ]
    }
    sens_gdf = gpd.GeoDataFrame({"geometry": [Point(0, 0)]}, crs="EPSG:32630")
    addr_gdf = gpd.GeoDataFrame(addr_points, crs="EPSG:32630")

    mask1_gdf = gpd.GeoDataFrame({"geometry": [Point(1, 0)]}, crs="EPSG:32630")
    results1 = analysis._calculate_k(
        sensitive_gdf=sens_gdf, candidate_gdf=mask1_gdf, address_gdf=addr_gdf
    )
    assert results1.loc[0, "k_anonymity"] == 2

    mask2_gdf = gpd.GeoDataFrame({"geometry": [Point(2, 0)]}, crs="EPSG:32630")
    results2 = analysis._calculate_k(
        sensitive_gdf=sens_gdf, candidate_gdf=mask2_gdf, address_gdf=addr_gdf
    )
    assert results2.loc[0, "k_anonymity"] == 4

    mask3_gdf = gpd.GeoDataFrame({"geometry": [Point(3, 0)]}, crs="EPSG:32630")
    results3 = analysis._calculate_k(
        sensitive_gdf=sens_gdf, candidate_gdf=mask3_gdf, address_gdf=addr_gdf
    )
    assert results3.loc[0, "k_anonymity"] == 5


def test_estimate_k_polygon():
    poly1 = Polygon([(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)])
    poly10 = Polygon([(0, 0), (1, 0), (1, -1), (-1, 0), (0, 0)])
    poly100 = Polygon([(0, 0), (0, -1), (-1, -1), (-1, 0), (0, 0)])
    poly1000 = Polygon([(0, 0), (-1, 0), (-1, 1), (0, 1), (0, 0)])
    census_poly = {
        "pop": [1, 10, 100, 1000],
        "geometry": [poly1, poly10, poly100, poly1000],
    }
    pop_gdf = gpd.GeoDataFrame(census_poly, crs="EPSG:32630")

    # uncertainty area includes entirety of all four areas, thus k equals sum of all population
    # across all four, minus one
    sens1_gdf = gpd.GeoDataFrame({"geometry": [Point(3, 0)]}, crs="EPSG:32630")
    mask1_gdf = gpd.GeoDataFrame({"geometry": [Point(0, 0)]}, crs="EPSG:32630")
    results1 = analysis._estimate_k(
        sensitive_gdf=sens1_gdf, candidate_gdf=mask1_gdf, population_gdf=pop_gdf
    )
    assert results1.loc[0, "k_anonymity"] == sum(census_poly["pop"]) - 1

    # uncertainty area only covers part of the 1000 pop area. As pop = 1000, and
    # coverage is bottom right quadrant of a buffer centered on top left corner
    # of top left quadrant, k should roughly equal ((population * pi * radius)/4) - 1
    sens2_gdf = gpd.GeoDataFrame({"geometry": [Point(0, 1)]}, crs="EPSG:32630")
    mask2_gdf = gpd.GeoDataFrame({"geometry": [Point(-1, 1)]}, crs="EPSG:32630")
    expected_k = (
        ((mask2_gdf.buffer(mask2_gdf.distance(sens2_gdf)).area * 1000) / 4).apply(floor)
    ) - 1

    results2 = analysis._estimate_k(
        sensitive_gdf=sens2_gdf, candidate_gdf=mask2_gdf, population_gdf=pop_gdf
    )
    assert results2.loc[0, "k_anonymity"] == expected_k[0]

    # Uncertainty area is a circle from the center at 0,0, with only partial
    # but equal coverage of each quadrant.
    sens3_gdf = gpd.GeoDataFrame({"geometry": [Point(1, 0)]}, crs="EPSG:32630")
    mask3_gdf = gpd.GeoDataFrame({"geometry": [Point(0, 0)]}, crs="EPSG:32630")
    results3 = analysis._estimate_k(
        sensitive_gdf=sens3_gdf, candidate_gdf=mask3_gdf, population_gdf=pop_gdf
    )

    area = mask3_gdf.buffer(mask3_gdf.distance(sens3_gdf)).area / 4
    expected_k = ((1 * area) + (10 * area) + (100 * area) + (1000 * area)).apply(floor) - 1
    assert results3.loc[0, "k_anonymity"] == expected_k[0]


def test_mean_center_drift(points):
    masked_points = points.copy(deep=True)
    masked_points["geometry"] = masked_points.geometry.translate(50, 0, 0)
    drift = analysis.central_drift(points, masked_points)
    assert drift == 50


def test_ripleys_k(points, tmpdir):
    masked = donut(points, 100, 500)
    kresult_sensitive = analysis.ripleys_k(points)
    kresult_masked = analysis.ripleys_k(masked)

    analysis.graph_ripleyresult(kresult_sensitive).savefig("SensitiveResult.png")
    assert os.path.exists("SensitiveResult.png")
    analysis.graph_ripleyresults(kresult_sensitive, kresult_masked, subtitle="Test Data").savefig(
        "ComparisonResult.png"
    )
    assert os.path.exists("ComparisonResult.png")


def test_ripleys_rmse(points):
    masked = donut(points, 1, 5)
    kresult_sensitive = analysis.ripleys_k(points)
    kresult_masked = analysis.ripleys_k(masked)
    rmse_1 = analysis.ripley_rmse(kresult_sensitive, kresult_masked)

    masked = donut(points, 1000, 5000)
    kresult_sensitive = analysis.ripleys_k(points)
    kresult_masked = analysis.ripleys_k(masked)
    rmse_2 = analysis.ripley_rmse(kresult_sensitive, kresult_masked)

    assert rmse_1 < rmse_2


def test_nearest_neighbor_stats(points):
    masked_points = points.copy()
    masked_points["geometry"] = masked_points.geometry.translate(50, 0, 0)
    nnd = analysis.nnd_delta(points, masked_points)
    assert nnd["nnd_min_delta"] == 0
    assert nnd["nnd_max_delta"] == 0
    assert nnd["nnd_mean_delta"] == 0

    masked_points = donut(points, 100, 500)
    nnd = analysis.nnd_delta(points, masked_points)
    assert isinstance(nnd["nnd_min_delta"], float)
    assert isinstance(nnd["nnd_max_delta"], float)
    assert isinstance(nnd["nnd_mean_delta"], float)


def test_map_displacement(points, tmpdir, address):
    masked_points = points.copy()
    masked_points["geometry"] = masked_points.geometry.translate(500, 0, 0)
    analysis.map_displacement(
        points, masked_points, filename="MapDisplacement.png", context_gdf=address
    )
    assert os.path.exists("MapDisplacement.png")


def test_evaluate(points, address):
    masked_points = points.copy(deep=True)
    masked_points["geometry"] = masked_points.geometry.translate(50, 0, 0)
    stats = analysis.evaluate(points, masked_points, address, skip_slow=False)

    assert stats["central_drift"] == 50
    assert stats["displacement_min"] == 50
    assert stats["k_max"] > stats["k_min"]
    assert stats["k_satisfaction_50"] == 0.0
    assert stats["nnd_min_delta"] == 0.0
    assert stats["ripley_rmse"] == 0.0



from dataclasses import dataclass
from math import sqrt

from geopandas import GeoDataFrame
from shapely import Point
from shapely.affinity import translate

from .. import tools


def donut(
    gdf: GeoDataFrame,
    low: float,
    high: float,
    container: GeoDataFrame = None,
    distribution: str = "uniform",
    seed: int = None,
    snap_to_streets: bool = False,
) -> GeoDataFrame:
    """
    Apply donut masking to a GeoDataFrame, randomly displacing points between a minimum and
    maximum distance. Advantages of this mask is speed and simplicity, though it does not
    handle highly varied population densities well.

    Example
    -------
    ```python
    from maskmypy import donut

    masked = donut(
        gdf=sensitive_points,
        min=100,
        max=1000
    )
    ```

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame containing sensitive points.
    low : float
        Minimum distance to displace points. Unit must match that of the `gdf` CRS.
    high : float
        Maximum displacement to displace points. Unit must match that of the `gdf` CRS.
    container : GeoDataFrame
        A  GeoDataFrame containing polygons within which intersecting sensitive points should
        remain after masking. This works by masking a point, checking if it intersects
        the same polygon prior to masking, and retrying until it does. Useful for preserving
        statistical relationships, such as census tract, or to ensure that points are not
        displaced into impossible locations, such as the ocean. CRS must match that of `gdf`.
    distribution : str
        The distribution used to determine masking distances. `uniform` provides
        a flat distribution where any value between the minimum and maximum distance is
        equally likely to be selected. `areal` is more likely to select distances that are
        further away. The `gaussian` distribution uses a normal distribution, where values
        towards the middle of the range are most likely to be selected. Note that gaussian
        distribution has a small chance of selecting values beyond the defined minimum and
        maximum.
    seed : int
        Used to seed the random number generator so that masked datasets are reproducible.
        Randomly generated if left undefined.
    snap_to_streets : bool
        If True, points are snapped to the nearest node on the OSM street network after masking.
        This can reduce the chance of false-attribution.

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame containing masked points.
    """
    _gdf = gdf.copy()
    _validate_donut(_gdf, low, high, container)

    seed = tools.gen_seed() if not seed else seed

    args = locals()
    del args["snap_to_streets"]
    del args["gdf"]

    masked_gdf = _Donut(**args).run()

    if snap_to_streets:
        masked_gdf = tools.snap_to_streets(masked_gdf)

    return masked_gdf


def _validate_donut(gdf, low, high, container):
    tools._validate_geom_type(gdf, "Point")

    if low >= high:
        raise ValueError("Minimum displacement distance is greater than or equal to maximum.")

    if container is not None:
        if not isinstance(container, GeoDataFrame):
            raise ValueError("Container is not a valid GeoDataFrame")
        tools._validate_geom_type(container, "Polygon", "MultiPolygon")
        tools._validate_crs(gdf.crs, container.crs)


@dataclass
class _Donut:
    _gdf: GeoDataFrame
    low: float
    high: float
    container: GeoDataFrame = None
    distribution: str = "uniform"
    seed: int = None

    def __post_init__(self) -> None:
        self._rng = tools.gen_rng(seed=self.seed)

    def _generate_random_offset(self) -> tuple[float, float]:
        if self.distribution == "uniform":
            hypotenuse = self._rng.uniform(self.low, self.high)
            x = self._rng.uniform(0, hypotenuse)
        elif self.distribution == "gaussian":
            mean = ((self.high - self.low) / 2) + self.low
            sigma = ((self.high - self.low) / 2) / 2.5
            hypotenuse = abs(self._rng.normal(mean, sigma))
            x = self._rng.uniform(0, hypotenuse)
        elif self.distribution == "areal":
            hypotenuse = 0
            while hypotenuse == 0:
                r1 = self._rng.uniform(self.low, self.high)
                r2 = self._rng.uniform(self.low, self.high)
                if r1 > r2:
                    hypotenuse = r1
            x = self._rng.uniform(0, hypotenuse)
        else:
            raise ValueError("Unknown distribution")

        y = sqrt(hypotenuse**2 - x**2)

        direction = self._rng.random()
        if direction < 0.25:
            x = x * -1
        elif direction < 0.5:
            y = y * -1
        elif direction < 0.75:
            x = x * -1
            y = y * -1
        elif direction < 1:
            pass
        return (x, y)

    def _mask_point(self, point: Point) -> Point:
        xoff, yoff = self._generate_random_offset()
        new_point = translate(point, xoff=xoff, yoff=yoff)
        return new_point

    def _mask_contained_point(self, point: Point) -> Point:
        intersection = self.container.intersects(point)
        intersected_polygons = set(
            intersection[intersection].index
        )  # Filter for rows where intersection == True, convert to set for later comparison
        start = intersected_polygons if len(intersected_polygons) > 0 else -1
        if len(start) > 1:
            raise ValueError(
                f"Point at {point} intersects {len(start)} polygons in the container. Container polygons must not overlap."
            )
        end = None
        while start != end:
            new_point = self._mask_point(point)
            intersection = self.container.intersects(new_point)
            intersected_polygons = set(intersection[intersection].index)
            end = intersected_polygons if len(intersected_polygons) > 0 else -1
        return new_point

    def run(self) -> GeoDataFrame:
        if isinstance(self.container, GeoDataFrame):
            self._gdf[self._gdf.geometry.name] = self._gdf[self._gdf.geometry.name].apply(
                self._mask_contained_point
            )
        else:
            self._gdf[self._gdf.geometry.name] = self._gdf[self._gdf.geometry.name].apply(
                self._mask_point
            )

        return self._gdf



from setuptools import find_packages, setup

with open("docs/README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="maskmypy",
    version="1.0.0",
    author="David Swanlund",
    author_email="maskmypy@swanlund.dev",
    description="Python tools for anonymizing geographic data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    install_requires=[
        "osmnx>=1.7.0,<2.0",
        "scikit-learn>=1.1.1,<2.0",
        "pointpats>=2.3.0,<3.0",
    ],
    extras_require={
        "develop": [
            "pytest",
            "black",
            "mkdocs-material",
            "mkdocs-roamlinks-plugin",
            "mkdocs-git-revision-date-localized-plugin",
            "mkdocstrings-python",
        ],
        "extra": ["contextily>=1.2.0"],
    },
    python_requires=">=3.10",
)



from math import sqrt

import matplotlib.pyplot as plt
from geopandas import GeoDataFrame, sjoin
from matplotlib.axis import Axis
from matplotlib.figure import Figure
from numpy import array, floor, square
from pointpats import PointPattern, k_test
from pointpats.distance_statistics import KtestResult
from shapely.geometry import LineString

from . import tools


def evaluate(
    sensitive_gdf: GeoDataFrame,
    candidate_gdf: GeoDataFrame,
    population_gdf: GeoDataFrame = None,
    population_column: str = "pop",
    skip_slow: bool = True,
) -> dict:
    """
    Evaluate the privacy protection and information loss of a masked dataset (`candidate_gdf`)
    compared to the unmasked sensitive dataset (`sensitive_gdf`). This is a convenience function
    that automatically runs many of the analysis tools that MaskMyPy offers, returning a simple
    dictionary of results. Note that privacy metrics require a `population_gdf` to be provided.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points to be evaluated.
    population_gdf : GeoDataFrame
        A GeoDataFrame containing either address points or polygons with a population column
        (see `population_column`). Used to calculate k-anonymity metrics.
    population_column : str
        If a polygon-based `population_gdf` is provided, the name of the column containing
        population counts.
    skip_slow : bool
        If True, skips analyses that are known to be slow. Currently, this only includes the
        root-mean-square error of Ripley's K results between the masked and unmasked data.

    Returns
    -------
    dict
        A dictionary containing evaluation results.
    """
    stats = {}

    # Information Loss
    stats["central_drift"] = central_drift(
        sensitive_gdf=sensitive_gdf, candidate_gdf=candidate_gdf
    )
    stats.update(
        summarize_displacement(
            displacement(
                sensitive_gdf=sensitive_gdf,
                candidate_gdf=candidate_gdf,
            )
        )
    )
    stats.update(nnd_delta(sensitive_gdf=sensitive_gdf, candidate_gdf=candidate_gdf))
    if not skip_slow:
        stats["ripley_rmse"] = ripley_rmse(ripleys_k(sensitive_gdf), ripleys_k(candidate_gdf))

    # Privacy
    if isinstance(population_gdf, GeoDataFrame):
        k_gdf = k_anonymity(
            sensitive_gdf=sensitive_gdf,
            candidate_gdf=candidate_gdf,
            population_gdf=population_gdf,
            population_column=population_column,
        )
        stats.update(summarize_k(k_gdf))
        stats["k_satisfaction_5"] = k_satisfaction(k_gdf, 5)
        stats["k_satisfaction_25"] = k_satisfaction(k_gdf, 25)
        stats["k_satisfaction_50"] = k_satisfaction(k_gdf, 50)
    return stats


def displacement(
    sensitive_gdf: GeoDataFrame, candidate_gdf: GeoDataFrame, col: str = "_distance"
) -> GeoDataFrame:
    """
    Adds a column to the `candidate_gdf` containing the distance between each masked point
    and its original, unmasked location (`sensitive_gdf`).

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points.
    col : str
        Name of the displacement distance column to add to `candidate_gdf`.

    Returns
    -------
    GeoDataFrame
        The `candidate_gdf` with an additional column describing displacement distance.
    """
    candidate_gdf = candidate_gdf.copy()
    candidate_gdf[col] = candidate_gdf.geometry.distance(sensitive_gdf.geometry)
    return candidate_gdf


def k_anonymity(
    sensitive_gdf: GeoDataFrame,
    candidate_gdf: GeoDataFrame,
    population_gdf: GeoDataFrame,
    population_column: str = "pop",
) -> GeoDataFrame:
    """
    Adds a column to the `candidate_gdf` containing the spatial k-anonymity value of each
    masked point.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points.
    population_gdf : GeoDataFrame
        A GeoDataFrame containing either address points or polygons with a population column
        (see `population_column`). Used to calculate k-anonymity metrics. Note that
        address points tend to provide more accurate results.
    population_column : str
        If a polygon-based `population_gdf` is provided, the name of the column containing
        population counts.

    Returns
    -------
    GeoDataFrame
        The `candidate_gdf` with an additional column describing k-anonymity.
    """
    if tools._validate_geom_type(population_gdf, "Point"):
        k_gdf = _calculate_k(sensitive_gdf, candidate_gdf, population_gdf)
    elif tools._validate_geom_type(population_gdf, "Polygon", "MultiPolygon"):
        if population_column not in population_gdf:
            raise ValueError(
                f"Cannot find population column {population_column} in population_gdf"
            )
        k_gdf = _estimate_k(sensitive_gdf, candidate_gdf, population_gdf, population_column)
    else:
        raise ValueError("population_gdf must include either Points or Polygons/MultiPolygons.")
    return k_gdf


def k_satisfaction(gdf: GeoDataFrame, min_k: int, col: str = "k_anonymity") -> float:
    """
    For a masked GeoDataFrame containing k-anonymity values, calculate the percentage of
    points that are equal to or greater than (i.e. satisfy) a given k-anonymity threshold (`min_k`).

    Parameters
    ----------
    gdf : GeoDataFrame
        A GeoDataFrame containing k-anonymity values.
    min_k : int
        The minimum k-anonymity that must be satisfied.
    col : str
        Name of the column containing k-anonymity values.

    Returns
    -------
    float
        A percentage of points in the GeoDataFrame that satisfy `min_k`.
    """
    return round(gdf.loc[gdf[col] >= min_k, col].count() / gdf[col].count(), 3)


def summarize_k(gdf: GeoDataFrame, col: str = "k_anonymity") -> dict:
    """
    For a masked GeoDataFrame containing k-anonymity values, calculate the minimum, maximum,
    median, and mean k-anonymity.

    Parameters
    ----------
    gdf : GeoDataFrame
        A GeoDataFrame containing k-anonymity values.
    col : str
        Name of the column containing k-anonymity values.

    Returns
    -------
    dict
        A dictionary containing summary k-anonymity statistics.
    """
    return {
        "k_min": int(gdf.loc[:, col].min()),
        "k_max": int(gdf.loc[:, col].max()),
        "k_med": round(float(gdf.loc[:, col].median()), 2),
        "k_mean": round(float(gdf.loc[:, col].mean()), 2),
    }


def summarize_displacement(gdf: GeoDataFrame, col: str = "_distance") -> dict:
    """
    For a masked GeoDataFrame containing displacement distances, calculate the minimum, maximum,
    median, and mean displacement distance.

    Parameters
    ----------
    gdf : GeoDataFrame
        A GeoDataFrame containing displacement distance values.
    col : str
        Name of the column containing displacement distance values.

    Returns
    -------
    dict
        A dictionary containing summary displacement distance statistics.
    """
    return {
        "displacement_min": round(float(gdf.loc[:, col].min()), 6),
        "displacement_max": round(float(gdf.loc[:, col].max()), 6),
        "displacement_med": round(float(gdf.loc[:, col].median()), 6),
        "displacement_mean": round(float(gdf.loc[:, col].mean()), 6),
    }


def nnd(gdf: GeoDataFrame) -> dict:
    """
    Calculate the minimum, maximum, and mean nearest neighbor distance for a given GeoDataFrame.

    Parameters
    ----------
    gdf : GeoDataFrame
        A GeoDataFrame containing points.

    Returns
    -------
    dict
        A dictionary containing the minimum, maximum, and mean nearest neighbor distance.
    """
    pp = _gdf_to_pointpattern(gdf)
    return {"nnd_min": pp.min_nnd, "nnd_max": pp.max_nnd, "nnd_mean": pp.mean_nnd}


def nnd_delta(sensitive_gdf: GeoDataFrame, candidate_gdf: GeoDataFrame) -> dict:
    """
    Calculate the *difference* between minimum, maximum, and mean nearest neighbor distances
    before (`sensitive_gdf`) and after (`candidate_gdf`) masking. Higher values indicate
    greater information loss due to masking.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points.

    Returns
    -------
    dict
        A dictionary describing deltas in nearest neighbor distance before and after masking.
    """
    before = nnd(sensitive_gdf)
    after = nnd(candidate_gdf)
    delta = {}
    for key, value in before.items():
        delta.update({f"{key}_delta": round(after[key] - before[key], 6)})
    return delta


def central_drift(sensitive_gdf: GeoDataFrame, candidate_gdf: GeoDataFrame) -> float:
    """
    Calculates how far the centroid of the point pattern has been displaced due to masking.
    Higher central drift indicates more information loss.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points.

    Returns
    -------
    float
        The central drift, with units equal to the CRS of the `sensitive_gdf`.
    """
    centroid_a = sensitive_gdf.dissolve().centroid
    centroid_b = candidate_gdf.dissolve().centroid
    return round(float(centroid_a.distance(centroid_b).iloc[0]), 6)


def ripleys_k(
    gdf: GeoDataFrame,
    max_dist: float = None,
    min_dist: float = None,
    steps: int = 10,
    simulations: int = 99,
) -> KtestResult:
    """
    Performs Ripley's K clustering analysis on a GeoDataFrame. This evaluates clustering across a
    range of spatial scales.

    See `maskmypy.analysis.ripley_rmse()`, `maskmypy.analysis.graph_ripleyresult()`, and
    `maskmypy.analysis.graph_ripleyresults()` for functions that process/visualize the results
    of this function.

    Parameters
    ----------
    gdf : GeoDataFrame
        GeoDataFrame to analyse.
    max_dist : float
        The largest distance band used for cluster analysis. If `None`, this defaults to one
        quarter of the smallest side of the bounding box (i.e. Ripleys Rule of Thumb).
    min_dist : float
        The smallest distance band used for cluster analysis. If `None`, this is automatically set
        to  `max_dist / steps`.
    steps : int
        The number of equally spaced intervals between the minimum and maximum distance bands
        to analyze clustering on.
    simulations : int
        The number of simulations to perform.

    Returns
    -------
    KtestResult
        A named tuple that contains `("support", "statistic", "pvalue", "simulations")`.
    """
    if not max_dist:
        max_dist = _gdf_to_pointpattern(gdf).rot

    if not min_dist:
        min_dist = max_dist / steps

    k_results = k_test(
        array(list(zip(gdf.geometry.x, gdf.geometry.y))),
        keep_simulations=True,
        support=(min_dist, max_dist, steps),
        n_simulations=simulations,
    )
    return k_results


def ripley_rmse(sensitive_result: KtestResult, candidate_result: KtestResult) -> float:
    """
    Calculates the root-mean-square error between the Ripley's K-test results of unmasked and
    masked data. As the goal of geographic masking is to reduce information loss, the actual
    amount of clustering in masked data is unimportant; what matters is that the clustering
    or dispersion of the masked data resembles that of the original, sensitive data. By comparing
    the RMSE of k-test results, we can reduce this deviation to a single figure, which is useful
    for quickly comparing how multiple masks perform.

    Lower RMSE values indicate less information loss due to masking, whereas higher values
    indicate greater information loss due to masking.

    Parameters
    ----------
    sensitive_result : KtestResult
        The KtestResult tuple from applying `maskmypy.analysis.ripleys_k()` on a sensitive layer.
    candidate_result : KtestResult
        The KtestResult tuple from applying `maskmypy.analysis.ripleys_k()` on a masked layer.

    Returns
    -------
    float
        The root-mean-square error between the two k-test results.
    """
    step_count = len(candidate_result.statistic)
    residuals = []
    for i in range(step_count):
        residual = candidate_result.statistic[i] - sensitive_result.statistic[i]
        residuals.append(residual)
    return round(sqrt(square(residuals).mean()), 3)


def graph_ripleyresult(result: KtestResult, subtitle: str = None) -> Figure:
    """
    Generate a graph depicting a given KtestResult, such as would be generated from using
    `maskmypy.analysis.ripleys_k()`.

    Parameters
    ----------
    result : KtestResult
        The KtestResult tuple from applying `maskmypy.analysis.ripleys_k()` on a given layer.
    subtitle : str
        A subtitle to add to the graph.

    Returns
    -------
    Figure
        A matplotlib.figure.Figure object.
    """
    bounds = _bounds_from_ripleyresult(result)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(result.support, bounds, color="#303030", label="Upper/Lower Bounds", alpha=0.25)
    ax.plot(result.support, result.statistic, color="#1f77b4", label="Observed K")
    ax.scatter(result.support, result.statistic, c="#1f77b4")
    ax.set_xlabel("Distance")
    ax.set_ylabel("K Function")
    ax.set_title(subtitle)
    _legend_deduped_labels(ax)
    fig.suptitle("K Function Plot")
    return fig


def graph_ripleyresults(
    sensitive_result: KtestResult,
    candidate_result: KtestResult,
    subtitle: str = None,
) -> Figure:
    """
    Generate a graph depicting two KtestResults, such as would be generated from using
    `maskmypy.analysis.ripleys_k()`.

    Similar to `maskmypy.analysis.graph_ripleyresult()` except this function graphs both
    the sensitive and candidate results, allowing for visual comparison of clustering and dispersion
    between the two.

    Parameters
    ----------
    sensitive_result : KtestResult
        The KtestResult tuple from applying `maskmypy.analysis.ripleys_k()` on the sensitive layer.
    candidate_result : KtestResult
        The KtestResult tuple from applying `maskmypy.analysis.ripleys_k()` on a masked layer.
    subtitle : str
        A subtitle to add to the graph.

    Returns
    -------
    Figure
        A matplotlib.figure.Figure object.
    """
    bounds = _bounds_from_ripleyresult(sensitive_result)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(
        sensitive_result.support,
        bounds,
        color="#ff7f0e",
        label="Sensitive Upper/Lower Bounds",
        alpha=0.35,
    )
    ax.plot(
        candidate_result.support,
        bounds,
        color="#1f77b4",
        label="Candidate Upper/Lower Bounds",
        alpha=0.35,
    )
    ax.plot(
        sensitive_result.support,
        sensitive_result.statistic,
        color="#ff7f0e",
        label="Sensitive Statistic",
    )
    ax.plot(
        candidate_result.support,
        candidate_result.statistic,
        color="#1f77b4",
        label="Candidate Statistic",
    )
    ax.scatter(sensitive_result.support, sensitive_result.statistic, zorder=6, c="#ff7f0e")
    ax.scatter(candidate_result.support, candidate_result.statistic, zorder=5, c="#1f77b4")
    ax.set_title(subtitle)
    ax.set_xlabel("Distance")
    ax.set_ylabel("K Function")
    _legend_deduped_labels(ax)
    fig.suptitle("K Function Result Comparison")
    return fig


def map_displacement(
    sensitive_gdf: GeoDataFrame,
    candidate_gdf: GeoDataFrame,
    filename: str = None,
    context_gdf: GeoDataFrame = None,
) -> plt:
    """
    Generate a map showing the displacement of each masked point from its original location.
    Requires the `contextily` package.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame containing sensitive points prior to masking.
    candidate_gdf : GeoDataFrame
        A GeoDataFrame containing masked points.
    filename : str
        If specified, saves the map to the filesystem.
    context_gdf : GeoDataFrame
        A GeoDataFrame containing contextual data to be added to the map, such as address points,
        administrative boundaries, etc.

    Returns
    -------
    matplotlib.pyplot
        A pyplot object containing the mapped data.
    """
    import contextily as ctx

    lines = sensitive_gdf.copy()
    lines = lines.join(candidate_gdf, how="left", rsuffix="_masked")
    lines.geometry = lines.apply(
        lambda x: LineString([x["geometry"], x["geometry_masked"]]), axis=1
    )
    ax = lines.plot(color="black", zorder=2, linewidth=1, figsize=[8, 8])
    ax = sensitive_gdf.plot(ax=ax, color="red", zorder=3, markersize=6)
    ax = candidate_gdf.plot(ax=ax, color="blue", zorder=4, markersize=6)
    if isinstance(context_gdf, GeoDataFrame):
        ax = context_gdf.plot(ax=ax, color="grey", zorder=1, markersize=3)

    ctx.add_basemap(ax, crs=sensitive_gdf.crs, source=ctx.providers.OpenStreetMap.Mapnik)
    plt.title("Displacement Distances", fontsize=16)
    plt.figtext(
        0.5,
        0.025,
        "Sensitive points (red), Masked points (blue). \n KEEP CONFIDENTIAL",
        wrap=True,
        horizontalalignment="center",
        fontsize=12,
    )
    if filename:
        plt.savefig(filename)

    return plt


def _disaggregate(gdf_a: GeoDataFrame, gdf_b: GeoDataFrame, gdf_b_col: str) -> GeoDataFrame:
    new_col = "_".join([gdf_b_col, "adjusted"])
    gdf_b["_b_area"] = gdf_b.geometry.area
    gdf = sjoin(gdf_a, gdf_b, how="left", rsuffix="b").rename_axis("_index_2").reset_index()
    gdf.geometry = gdf.apply(
        lambda x: x.geometry.intersection(gdf_b.at[x["index_b"], "geometry"]),
        axis=1,
    )
    gdf["_intersected_area"] = gdf.geometry.area

    for i, _ in enumerate(gdf.index):
        fragments = gdf.loc[gdf["_index_2"] == i, :]
        for index, row in fragments.iterrows():
            area_pct = row["_intersected_area"] / row["_b_area"]
            gdf.at[index, new_col] = row[gdf_b_col] * area_pct
    return gdf


def _gdf_to_pointpattern(gdf: GeoDataFrame) -> PointPattern:
    return PointPattern(list(zip(gdf.geometry.x, gdf.geometry.y)))


def _bounds_from_ripleyresult(result: KtestResult) -> list:
    step_count = len(result.simulations[0])
    lower_bounds = []
    upper_bounds = []
    for i in range(step_count):
        values = [sim[i] for sim in result.simulations]
        lower_bounds.append(min(values))
        upper_bounds.append(max(values))
    return list(zip(lower_bounds, upper_bounds))


def _legend_deduped_labels(ax: Axis) -> None:
    handles, labels = ax.get_legend_handles_labels()
    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
    ax.legend(*zip(*unique))


def _estimate_k(
    sensitive_gdf: GeoDataFrame,
    candidate_gdf: GeoDataFrame,
    population_gdf: GeoDataFrame,
    population_column: str = "pop",
) -> GeoDataFrame:
    candidate_gdf = candidate_gdf.copy()
    pop_col_adjusted = "_".join([population_column, "adjusted"])
    candidate_gdf["k_anonymity"] = (
        displacement(sensitive_gdf, candidate_gdf)
        .assign(geometry=lambda x: x.geometry.buffer(x["_distance"]), axis=1)
        .pipe(_disaggregate, gdf_b=population_gdf, gdf_b_col=population_column)
        .groupby("_index_2")[pop_col_adjusted]
        .sum()
        .apply(floor)
    ) - 1
    return candidate_gdf


def _calculate_k(
    sensitive_gdf: GeoDataFrame, candidate_gdf: GeoDataFrame, address_gdf: GeoDataFrame
) -> GeoDataFrame:
    candidate_gdf = candidate_gdf.copy()
    candidate_gdf_tmp = displacement(sensitive_gdf, candidate_gdf).assign(
        geometry=lambda x: x.buffer(x["_distance"])
    )
    candidate_gdf["k_anonymity"] = (
        sjoin(address_gdf, candidate_gdf_tmp, how="left", rsuffix="candidate")
        .groupby("index_candidate")
        .size()
    )
    candidate_gdf.fillna({"k_anonymity": 0}, inplace=True)
    return candidate_gdf
